import datetime
import pathlib
import re
import subprocess
from dataclasses import (
    dataclass,
    field,
)
from typing import (
    Dict,
    List,
    Optional,
    Set,
)

import click
import docutils
from docutils import (
    frontend,
    utils,
)
from docutils.parsers.rst import Parser
from github import PullRequest
from packaging.version import Version

from .cli.options import (
    ClickVersion,
    galaxy_root_option,
    group_options,
)
from .github_client import github_client
from .metadata import (
    _text_target,
    strip_release,
)

g = github_client()
PROJECT_OWNER = "galaxyproject"
PROJECT_NAME = "galaxy"
REPO = f"{PROJECT_OWNER}/{PROJECT_NAME}"
DEFAULT_UPSTREAM_URL = f"https://github.com/{REPO}.git"

HISTORY_TEMPLATE = """History
-------

.. to_doc

"""
RELEASE_BRANCH_REGEX = re.compile(r"^release_(\d{2}\.\d{1,2})$")
FIRST_RELEASE_CHANGELOG_TEXT = "First release"


@dataclass
class ReleaseItem:
    version: Version
    date: Optional[str] = None


@dataclass
class ChangelogItem:
    version: Version
    changes: List[str]
    date: Optional[str] = None

    @property
    def datetime_date(self) -> datetime.datetime:
        if self.date:
            return datetime.datetime.fromisoformat(self.date)
        raise Exception(f"Changelog item for version {self.version} is missing date.")

    @property
    def is_empty_devrelease(self):
        return (self.version.is_devrelease or self.version.is_prerelease) and not self.changes

    def __str__(self) -> str:
        change_lines = "\n".join(self.changes)
        if self.date:
            version_line = f"{self.version} ({self.date})"
        else:
            version_line = str(self.version)
        return f"""{'-' * len(version_line)}
{version_line}
{'-' * len(version_line)}

{change_lines}
"""


@dataclass
class Package:
    path: pathlib.Path
    current_version: str
    commits: Set[str] = field(default_factory=set)
    prs: Set[PullRequest.PullRequest] = field(default_factory=set)
    modified_paths: List[pathlib.Path] = field(default_factory=list)
    package_history: List[ChangelogItem] = field(default_factory=list)
    release_items: List[ReleaseItem] = field(default_factory=list)

    @property
    def name(self) -> str:
        return self.path.name

    @property
    def setup_cfg(self) -> pathlib.Path:
        return self.path / "setup.cfg"

    @property
    def history_rst(self) -> pathlib.Path:
        return self.path / "HISTORY.rst"

    @property
    def changelog(self) -> str:
        changelog_string = "\n".join(str(h) for h in self.package_history)
        return f"{HISTORY_TEMPLATE}{changelog_string}"

    def write_history(self):
        self.history_rst.write_text(self.changelog)
        self.modified_paths.append(self.history_rst)

    def add_release(self, item: ReleaseItem) -> None:
        self.release_items.append(item)

    @property
    def is_new(self) -> bool:
        """Package has not been released:
        - has only one release item,
        - which is a devrelease,
        - and has no date.
        """
        return (
            len(self.release_items) == 1
            and self.release_items[0].version.is_devrelease
            and not self.release_items[0].date
        )

    def __repr__(self) -> str:
        pretty_string = f"[Package: {self.name}, Current Version: {self.current_version}"
        return pretty_string


def get_sorted_package_paths(galaxy_root: pathlib.Path) -> List[pathlib.Path]:
    root_package_path = galaxy_root.joinpath("packages")
    sorted_packages = root_package_path.joinpath("packages_by_dep_dag.txt").read_text().splitlines()
    # Ignore empty lines and lines beginning with "#"
    return [root_package_path.joinpath(p) for p in sorted_packages if p and not p.startswith("#")]


def read_package(package_path: pathlib.Path) -> Package:
    setup_cfg = package_path / "setup.cfg"
    package = None
    with setup_cfg.open() as content:
        for line in content:
            if line.startswith("version = "):
                version = line.strip().split("version = ")[-1]
                package = Package(path=package_path, current_version=version)
                break
    if not package:
        raise ValueError(f"{setup_cfg} does not contain version line")
    package.package_history = parse_changelog(package)
    return package


def parse_changelog(package: Package) -> List[ChangelogItem]:
    def add_changelog_item(changes, child):
        rawsource = child.rawsource.strip()
        if rawsource:
            changes.append(f"* {rawsource}")

    settings = frontend.get_default_settings(Parser)  # type: ignore[attr-defined] ## upstream type stubs not updated?
    document = utils.new_document(str(package.history_rst), settings)
    Parser().parse(package.history_rst.read_text(), document)
    changelog_items: List[ChangelogItem] = []
    root_section = document[0]
    assert isinstance(root_section, docutils.nodes.section)
    for node in root_section.children:
        # ignore title and comment
        assert isinstance(node, (docutils.nodes.title, docutils.nodes.comment, docutils.nodes.section)), node
        if isinstance(node, docutils.nodes.section):
            release_version = node[0].astext()
            current_date = None
            if " (" in release_version:
                version_str, current_date = release_version.split(" (")
                current_date = current_date.split(")")[0]
            else:
                # this should be a dev-release without changelog items
                # we will just omit this later
                version_str = release_version
            current_version = Version(version_str)
            package.add_release(ReleaseItem(version=current_version, date=current_date))
            changes: List = []
            for changelog_item in node[1:]:
                # could be bullet list or a nested section with bugfix, docs, etc
                if isinstance(changelog_item, docutils.nodes.bullet_list):
                    for child in changelog_item.children:
                        add_changelog_item(changes, child)
                elif isinstance(changelog_item, docutils.nodes.paragraph):
                    changes = changelog_item.rawsource.splitlines()
                elif isinstance(changelog_item, docutils.nodes.section):
                    kind = changelog_item[0].astext()
                    section_delimiter = "=" * len(kind)
                    changes.append(f"\n{section_delimiter}\n{kind}\n{section_delimiter}\n")
                    for section_changelog_item in changelog_item[1:]:
                        if isinstance(section_changelog_item, docutils.nodes.system_message):
                            # Likely a warning that subsection (e.g. Bug fixes) is not unique
                            continue
                        assert isinstance(section_changelog_item, docutils.nodes.bullet_list), type(
                            section_changelog_item
                        )
                        for child in section_changelog_item:
                            add_changelog_item(changes, child)
            changelog_items.append(ChangelogItem(version=current_version, date=current_date, changes=changes))

    # Filter out dev release versions without changelog,
    # we're going to add these back after committing the release version
    clean_changelog_items: List[ChangelogItem] = []
    for item in changelog_items:
        if not (item.is_empty_devrelease or package.is_new):
            if item.date is None:
                raise Exception(
                    f"Error in '{package.history_rst}'. Changelog entry for non-dev version '{item.version}' has no date but contains changes. You have to fix this manually."
                )
            clean_changelog_items.append(item)
    return sorted(
        clean_changelog_items,
        key=lambda item: item.datetime_date,
        reverse=True,
    )


def bump_package_version(package: Package, new_version: Version):
    new_content = []
    content = package.setup_cfg.read_text().splitlines()
    for line in content:
        if line.startswith("version = "):
            line = f"version = {new_version}"
        new_content.append(line)
    package.setup_cfg.write_text("\n".join(new_content))
    package.modified_paths.append(package.setup_cfg)


def get_commits_since_last_version(package: Package, last_version_tag: str):
    click.echo(f"finding commits to package {package.name} made since {last_version_tag}")
    package_source_paths = []
    commits = set()
    for code_dir in ["galaxy", "tests", "galaxy_test"]:
        package_code_path = package.path / code_dir
        if package_code_path.exists():
            # get all symlinks pointing to a directory
            for item in package_code_path.iterdir():
                # Check if the item is a symlink and if its target points to a directory
                if item.is_symlink() and item.resolve().is_dir():
                    package_source_paths.append(item.resolve())
    for package_source_path in package_source_paths:
        result = subprocess.run(
            [
                "git",
                "log",
                "--oneline",
                "--no-merges",
                "--pretty=format:%h",
                f"{last_version_tag}..HEAD",
                package_source_path,
            ],
            cwd=package.path,
            capture_output=True,
            text=True,
        )
        result.check_returncode()
        for line in result.stdout.splitlines():
            if line:
                commits.add(line)
    package.commits = commits


def commits_to_prs(packages: List[Package]):
    commits = set.union(*(p.commits for p in packages))
    pr_cache = {}
    commit_to_pr = {}
    repo = g.get_repo(REPO)
    total_commits = len(commits)
    for i, commit in enumerate(commits):
        click.echo(f"Processing commit {i} of {total_commits}")
        # Get the list of pull requests associated with the commit
        commit_obj = repo.get_commit(commit)
        prs = commit_obj.get_pulls()
        if not prs:
            raise Exception(f"commit {commit} has no associated PRs")
        for pr in prs:
            if pr.number not in pr_cache:
                pr_cache[pr.number] = pr
            commit_to_pr[commit] = pr_cache[pr.number]
    for package in packages:
        # Exclude commits without PRs
        package.prs = set(commit_to_pr[commit] for commit in package.commits if commit in commit_to_pr)


def update_package_history(package: Package, new_version: Version):
    sorted_and_formatted_changes = []
    changes: Dict[str, List[str]] = {
        "Bug fixes": [],
        "Enhancements": [],
        "Other changes": [],
    }
    if package.is_new:
        # For new packages, replace any current text; do not list PRs.
        sorted_and_formatted_changes.append(FIRST_RELEASE_CHANGELOG_TEXT)
    elif not package.prs:
        # Skip publishing packages if no change ?
        sorted_and_formatted_changes.append("No recorded changes since last release")
    else:
        for pr in sorted(package.prs, key=lambda pr: pr.number):
            category = "Other changes"
            text_target = _text_target(pr, skip_merge=False)
            if text_target:
                if "bug" in text_target:
                    category = "Bug fixes"
                if "enhancement" in text_target or "feature" in text_target:
                    category = "Enhancements"
            changes[category].append(
                f"* {strip_release(pr.title)} by `@{pr.user.login} <https://github.com/{pr.user.login}>`_ in `#{pr.number} <{pr.html_url}>`_"
            )

    for kind, entries in changes.items():
        if entries:
            section_delimiter = "=" * len(kind)
            sorted_and_formatted_changes.append(f"\n{section_delimiter}\n{kind}\n{section_delimiter}\n")
        sorted_and_formatted_changes.extend(entries)

    now = datetime.datetime.now().strftime("%Y-%m-%d")
    package.package_history.insert(
        0,
        ChangelogItem(version=new_version, changes=sorted_and_formatted_changes, date=now),
    )
    package.write_history()


def build_package(package: Package):
    click.echo(f"Running make clean for package {package.name}")
    subprocess.run(["make", "clean"], cwd=package.path).check_returncode()
    click.echo(f"running make dist for package {package.name}")
    subprocess.run(["make", "dist"], cwd=package.path).check_returncode()
    click.echo(f"running make lint-dist for package {package.name}")
    subprocess.run(["make", "lint-dist"], cwd=package.path).check_returncode()


def upload_package(package: Package):
    click.echo(f"uploading package {package.name}")
    subprocess.run(
        ["twine", "upload", "--skip-existing"]
        + [str(artifact_path) for artifact_path in package.path.joinpath("dist").glob("*")],
        cwd=package.path,
    ).check_returncode()


def get_root_version(galaxy_root: pathlib.Path) -> Version:
    version_py = galaxy_root / "lib" / "galaxy" / "version.py"
    version_py_contents = version_py.read_text().splitlines()
    assert len(version_py_contents) == 3
    major_version = version_py_contents[0].split('"')[1]
    minor_version = version_py_contents[1].split('"')[1]
    return Version(f"{major_version}.{minor_version}")


def set_root_version(galaxy_root: pathlib.Path, new_version: Version) -> pathlib.Path:
    major_galaxy_release_string = f"{new_version.major}.{new_version.minor}"
    minor_galaxy_release_string = str(new_version).replace(f"{major_galaxy_release_string}.", "")
    VERSION_PY_TEMPLATE = f"""VERSION_MAJOR = "{major_galaxy_release_string}"
VERSION_MINOR = "{minor_galaxy_release_string}"
VERSION = VERSION_MAJOR + (f".{{VERSION_MINOR}}" if VERSION_MINOR else "")
"""
    version_py = galaxy_root / "lib" / "galaxy" / "version.py"
    version_py.write_text(VERSION_PY_TEMPLATE)
    return version_py


def is_git_clean(galaxy_root: pathlib.Path):
    click.echo(f"Making sure galaxy clone at '{galaxy_root}' is clean:")
    command = ["git", "diff-index", "--quiet", "HEAD"]
    result = subprocess.run(command, capture_output=True, text=True, cwd=galaxy_root)
    if result.returncode == 0:
        return True
    else:
        msg = f"Command '{' '.join(command)}' failed with exit code {result.returncode}"
        if result.returncode != 1 and result.stderr:
            msg = f"{msg}, stderr: {result.stderr}"
        click.echo(msg)
        return False


def get_current_branch(galaxy_root):
    current_branch_cmd = ["git", "rev-parse", "--abbrev-ref", "HEAD"]
    result = subprocess.run(current_branch_cmd, cwd=galaxy_root, capture_output=True, text=True)
    result.check_returncode()
    return result.stdout.strip()


def get_branches(galaxy_root: pathlib.Path, new_version: Version, current_branch: str):
    """
    Tries to get release and dev branches that we need to merge forward to.
    """
    major_minor_new_version = Version(f"{new_version.major}.{new_version.minor}")
    cmd = ["git", "branch", "--format=%(refname:short)"]
    result = subprocess.run(cmd, cwd=galaxy_root, capture_output=True, text=True)
    result.check_returncode()
    release_branches = []
    branches = result.stdout.splitlines()
    for branch in branches:
        match = RELEASE_BRANCH_REGEX.match(branch)
        if match:
            version = match.group(1)
            branch_version = Version(version)
            if branch_version > major_minor_new_version:
                release_branches.append(branch)
    if current_branch != "dev":
        release_branches.append("dev")
    return release_branches


def merge_and_resolve_branches(
    galaxy_root: pathlib.Path,
    base_branch: str,
    new_branch: str,
    packages: List[Package],
):
    checkout_cmd = ["git", "checkout", new_branch]
    subprocess.run(checkout_cmd, cwd=galaxy_root).check_returncode()

    package_paths = {p.path: p for p in packages}
    packages_to_rewrite: List[Package] = []
    for package_path in get_sorted_package_paths(galaxy_root):
        if package_path not in package_paths:
            continue
        package = read_package(package_path)
        packages_to_rewrite.append(package)

    merge_cmd = ["git", "merge", base_branch]
    result = subprocess.run(merge_cmd, cwd=galaxy_root, capture_output=True, text=True)
    merge_conflict = result.returncode != 0  # merge conflict expected
    # restore base galaxy version
    subprocess.run(
        [
            "git",
            "checkout",
            new_branch,
            str(galaxy_root / "lib" / "galaxy" / "version.py"),
        ],
        cwd=galaxy_root,
    )
    # we rewrite the packages changelog
    for new_package in packages_to_rewrite:
        previous_package = package_paths[new_package.path]
        combined_history = previous_package.package_history + new_package.package_history
        new_history: List[ChangelogItem] = []
        last_changelog_item: Optional[ChangelogItem] = None
        for changelog_item in sorted(combined_history, key=lambda item: item.version, reverse=True):
            if last_changelog_item and changelog_item.version == last_changelog_item.version:
                assert (
                    last_changelog_item.changes == changelog_item.changes
                ), f"Changelog differs for version {changelog_item.version} of package {new_package.name}, you have to fix this manually.\nOffending lines are {last_changelog_item.changes} and {changelog_item.changes}"
                continue
            if not changelog_item.date and not changelog_item.changes:
                # dev0 version, we'll inject that later
                continue
            new_history.append(changelog_item)
            last_changelog_item = changelog_item
        # we change the original package in place so this continues to work if we merge forward across multiple branches
        previous_package.package_history = sorted(
            new_history,
            key=lambda item: item.datetime_date,
            reverse=True,
        )
        dev_version = get_root_version(galaxy_root)
        previous_package.package_history.insert(0, ChangelogItem(version=dev_version, changes=[], date=None))
        previous_package.write_history()
        subprocess.run(["git", "add", str(previous_package.history_rst)], cwd=galaxy_root)
        # restore setup.cfg
        subprocess.run(
            ["git", "checkout", new_branch, str(previous_package.setup_cfg)], cwd=galaxy_root
        ).check_returncode()
    # Commit changes
    if merge_conflict:
        subprocess.run(["git", "commit", "--no-edit"], cwd=galaxy_root).check_returncode()
    else:
        subprocess.run(["git", "commit", "--amend", "--no-edit"], cwd=galaxy_root).check_returncode()


def get_next_devN_version(galaxy_root) -> Version:
    root_version = get_root_version(galaxy_root)
    if root_version.dev is not None:
        minor_version = root_version.minor
        micro_version = root_version.micro
        dev_version = root_version.dev + 1
    else:
        minor_version = root_version.minor
        micro_version = root_version.micro + 1
        dev_version = 0
    return Version(f"{root_version.major}.{minor_version}.{micro_version}.dev{dev_version}")


def is_merge_required(base_branch: str, new_branch: str, galaxy_root: pathlib.Path):
    subprocess.run(["git", "checkout", new_branch], cwd=galaxy_root).check_returncode()
    process = subprocess.run(
        ["git", "merge", "--no-commit", "--no-ff", base_branch],
        cwd=galaxy_root,
        capture_output=True,
    )
    if not process.stdout == b"Already up to date.\n":
        subprocess.run(["git", "merge", "--abort"], cwd=galaxy_root)
    if process.returncode == 0:
        return False
    return True


def ensure_branches_up_to_date(branches: List[str], base_branch: str, upstream: str, galaxy_root: pathlib.Path):
    for branch in branches:
        subprocess.run(["git", "checkout", branch], cwd=galaxy_root).check_returncode()
        # Check that the head commit matches the commit for the same branch at the specified remote repo url
        result = subprocess.run(
            ["git", "ls-remote", upstream, f"refs/heads/{branch}"],
            cwd=galaxy_root,
            capture_output=True,
            text=True,
        )
        result.check_returncode()
        remote_commit_hash = result.stdout.split("\t")[0]
        result = subprocess.run(["git", "rev-parse", "HEAD"], cwd=galaxy_root, capture_output=True, text=True)
        result.check_returncode()
        local_commit_hash = result.stdout.strip()
        if remote_commit_hash != local_commit_hash:
            raise Exception(
                f"Local tip of branch {branch} is {local_commit_hash}, remote tip of branch is {remote_commit_hash}. Make sure that your local branches are up to date and track f{upstream}."
            )
    subprocess.run(["git", "checkout", base_branch], cwd=galaxy_root).check_returncode()


def push_references(
    references: List[str], galaxy_root: pathlib.Path, upstream: str = "https://github.com/galaxyproject/galaxy.git"
):
    for reference in references:
        subprocess.run(["git", "push", upstream, reference], cwd=galaxy_root).check_returncode()


@click.group(help="Subcommands of this script can create new releases and build and upload package artifacts")
def cli():
    pass


packages_option = click.option(
    "--packages",
    "package_subset",
    multiple=True,
    type=str,
    default=None,
    help="Restrict build to specified packages",
)
no_confirm_option = click.option("--no-confirm", type=bool, is_flag=True, default=False)


@group_options(galaxy_root_option, packages_option, no_confirm_option)
@cli.command("build-and-upload", help="Build and upload packages")
def build_and_upload(
    galaxy_root: pathlib.Path,
    package_subset: List[str],
    no_confirm: bool,
):
    packages: List[Package] = []
    for package_path in get_sorted_package_paths(galaxy_root):
        if package_subset and package_path.name not in package_subset:
            continue
        package = read_package(package_path)
        packages.append(package)
    for package in packages:
        click.echo(f"Building package {package}")
        build_package(package)
    if no_confirm or click.confirm("Upload packages ?"):
        for package in packages:
            click.echo(f"Uploading package {package}")
            upload_package(package)


@cli.command("create-release", help="Create a new point release")
@group_options(galaxy_root_option)
@click.option(
    "--new-version",
    type=ClickVersion(),
    help="Specify new release version. Must be valid PEP 440 version",
)
@click.option(
    "--last-commit",
    type=str,
    help="Specify commit or tag that was used for the last package release. This is used to find the changelog for packages.",
)
@click.option("--build-packages/--no-build-packages", type=bool, is_flag=True, default=True)
@click.option("--upload-packages", type=bool, is_flag=True, default=False)
@click.option("--upstream", type=str, default=DEFAULT_UPSTREAM_URL)
@group_options(packages_option, no_confirm_option)
def create_point_release(
    galaxy_root: pathlib.Path,
    new_version: Version,
    build_packages: bool,
    last_commit: str,
    package_subset: List[str],
    upload_packages: bool,
    no_confirm: bool,
    upstream: str,
):
    # Update version.py
    if not is_git_clean(galaxy_root):
        click.confirm(
            "Your galaxy clone has untracked or staged changes, are you sure you want to continue ?",
            abort=True,
        )
    root_version = get_root_version(galaxy_root)
    base_branch = current_branch = get_current_branch(galaxy_root)
    click.echo(
        f"- Current Galaxy version: {root_version}\n- New Galaxy version: {new_version}\n- Base branch: {base_branch}"
    )
    if not no_confirm:
        click.confirm("Does this look correct?", abort=True)
    newer_branches = get_branches(galaxy_root, new_version, base_branch)
    all_branches = newer_branches + [base_branch]
    click.echo("Making sure that all branches are up to date")
    ensure_branches_up_to_date(all_branches, base_branch, upstream, galaxy_root)

    click.echo("Making sure that merging forward will result in clean merges")
    for new_branch in newer_branches:
        merge_required = is_merge_required(base_branch=current_branch, new_branch=new_branch, galaxy_root=galaxy_root)
        if merge_required:
            msg = f"Merge conflicts occurred while attempting to merge branch {current_branch} into {new_branch}. You should resolve conflicts and try again."
            if no_confirm:
                raise Exception(msg)
            click.echo(msg)
            if click.confirm("Continue anyway ?", abort=True):
                current_branch = new_branch
                break
    subprocess.run(["git", "checkout", base_branch], cwd=galaxy_root).check_returncode()
    modified_paths = [set_root_version(galaxy_root, new_version)]
    # read packages and find prs that affect a package
    packages: List[Package] = []
    for package_path in get_sorted_package_paths(galaxy_root):
        if package_subset and package_path.name not in package_subset:
            continue
        package = read_package(package_path)
        packages.append(package)
        get_commits_since_last_version(package, last_commit)
    commits_to_prs(packages)
    # update package versions and changelog files
    for package in packages:
        if new_version:
            bump_package_version(package, new_version)
            update_package_history(package, new_version)
        if build_packages:
            build_package(package)
        modified_paths.extend(package.modified_paths)
    # show changed paths, optionally run git diff
    changed_paths = [str(p) for p in modified_paths]
    pretty_paths = "\n".join(changed_paths)
    click.echo(f"The following paths have been modified: \n{pretty_paths}")
    if not no_confirm and click.confirm("show diff ?"):
        cmd = ["git", "diff"]
        cmd.extend([str(p) for p in modified_paths])
        subprocess.run(cmd, cwd=galaxy_root)
    if build_packages and upload_packages and (no_confirm or click.confirm("Upload packages to ?")):
        for package in packages:
            upload_package(package)
    # stage changes, commit and tag
    if not no_confirm:
        click.confirm("Stage and commit changes ?", abort=True)
    cmd = ["git", "add"]
    cmd.extend(changed_paths)
    release_tag = f"v{new_version}"
    subprocess.run(cmd, cwd=galaxy_root)
    subprocess.run(["git", "commit", "-m" f"Create version {new_version}"], cwd=galaxy_root)
    if not no_confirm:
        click.confirm(f"Create git tag '{release_tag}'?", abort=True)

    subprocess.run(["git", "tag", release_tag], cwd=galaxy_root)
    dev_version = get_next_devN_version(galaxy_root)
    version_py = set_root_version(galaxy_root, dev_version)
    modified_paths = [version_py]
    for package in packages:
        bump_package_version(package, dev_version)
        package.package_history.insert(0, ChangelogItem(version=dev_version, changes=[], date=None))
        package.write_history()
        modified_paths.extend(package.modified_paths)
    cmd = ["git", "add"]
    cmd.extend([str(p) for p in modified_paths])
    subprocess.run(cmd, cwd=galaxy_root)
    subprocess.run(["git", "commit", "-m", f"Start work on {dev_version}"], cwd=galaxy_root)
    # merge changes into newer branches
    # special care needs to be taken for changelog files
    if not no_confirm and newer_branches:
        click.confirm(
            f"Merge branch '{base_branch}' into {', '.join(newer_branches)} ?",
            abort=True,
        )
    current_branch = base_branch
    for new_branch in newer_branches:
        click.echo(f"Merging {base_branch} into {new_branch}")
        merge_and_resolve_branches(galaxy_root, current_branch, new_branch, packages)
        current_branch = new_branch
    references = [release_tag] + all_branches
    if no_confirm or click.confirm(f"Push {','.join(references)} to upstream '{upstream}' ?", abort=True):
        push_references(references=references, galaxy_root=galaxy_root, upstream=upstream)


if __name__ == "__main__":
    cli()



# Little script to make HISTORY.rst more easy to format properly, lots TODO
# pull message down and embed, handle multiple, etc...

import calendar
import datetime
import os
import re
import string
import sys
import textwrap
from pathlib import Path
from typing import Optional

import click
from github.PullRequest import PullRequest
from packaging.version import Version

from .cli.options import (
    ClickVersion,
    galaxy_root_option,
    group_options,
)
from .github_client import github_client
from .metadata import (
    _pr_to_labels,
    _pr_to_str,
    _text_target,
    GROUPED_TAGS,
    PROJECT_NAME,
    PROJECT_OWNER,
    PROJECT_URL,
    strip_release,
)

RELEASE_DELTA_MONTHS = 4  # Number of months between releases.
MINOR_TO_MONTH = {0: 2, 1: 6, 2: 10}


TEMPLATE = """
.. to_doc

${release}
===============================

.. announce_start

Enhancements
-------------------------------

.. major_feature


.. feature

.. enhancement

.. small_enhancement



Fixes
-------------------------------

.. major_bug


.. bug


.. include:: ${release}_prs.rst

"""

ANNOUNCE_TEMPLATE = string.Template(
    """
===========================================================
${month_name} 20${year} Galaxy Release (v ${release})
===========================================================

.. include:: _header.rst

Highlights
===========================================================

Feature1
--------

Feature description.

Feature2
--------

Feature description.

Feature3
--------

Feature description.

Also check out the `${release} user release notes <${release}_announce_user.html>`__.
Are you an admin? Check out `some admin relevant PRs <https://github.com/galaxyproject/galaxy/pulls?q=label%3Ahighlight%2Fadmin+milestone%3A${release}+is%3Aclosed+is%3Apr>`__.

Get Galaxy
===========================================================

The code lives at `GitHub <https://github.com/galaxyproject/galaxy>`__ and you should have `Git <https://git-scm.com/>`__ to obtain it.

To get a new Galaxy repository run:
  .. code-block:: shell

      $$ git clone -b release_${release} https://github.com/galaxyproject/galaxy.git

To update an existing Galaxy repository run:
  .. code-block:: shell

      $$ git fetch origin && git checkout release_${release} && git pull --ff-only origin release_${release}

See the `community hub <https://galaxyproject.org/develop/source-code/>`__ for additional details on source code locations.


Administration Notes
===========================================================
Add content or drop section.

Configuration Changes
===========================================================
Add content or drop section.

Deprecation Notices
===========================================================
Add content or drop section.

Release Notes
===========================================================

.. include:: ${release}.rst
   :start-after: announce_start

.. include:: _thanks.rst
"""  # noqa: E501
)

ANNOUNCE_USER_TEMPLATE = string.Template(
    """
===========================================================
${month_name} 20${year} Galaxy Release (v ${release})
===========================================================

.. include:: _header.rst

Highlights
===========================================================

Feature1
--------

Feature description.

Feature2
--------

Feature description.

Feature3
--------

Feature description.


Visualizations
===========================================================

.. visualizations

Datatypes
===========================================================

.. datatypes

Builtin Tool Updates
===========================================================

.. tools

Release Testing Team
===========================================================

A special thanks to the release testing team for testing many of the new features and reporting many bugs:

<team members go here>

Release Notes
===========================================================

Please see the :doc:`full release notes <${release}_announce>` for more details.

.. include:: ${release}_prs.rst

.. include:: _thanks.rst
"""  # noqa: E501
)

NEXT_TEMPLATE = string.Template(
    """
:orphan:

===========================================================
${month_name} 20${year} Galaxy Release (v ${version})
===========================================================


Schedule
===========================================================
 * Planned Freeze Date: ${freeze_date}
 * Planned Release Date: ${release_date}
"""
)

PRS_TEMPLATE = """
.. github_links
"""

RELEASE_ISSUE_TEMPLATE = string.Template(
    """

- [X] **Prep**

    - [X] ~~Create this release issue ``make release-issue``.~~
    - [X] ~~Set freeze date (${freeze_date}).~~

- [ ] **Branch Release (on or around ${freeze_date})**

    - [ ] Ensure all [blocking milestone pull requests](https://github.com/galaxyproject/galaxy/pulls?q=is%3Aopen+is%3Apr+milestone%3A${version}) have been merged, delayed, or closed.

          make release-check-blocking-prs

    - [ ] Add latest database revision identifier (for ``release_${version}`` and ``${version}``) to ``REVISION_TAGS`` in ``galaxy/model/migrations/dbscript.py``.

    - [ ] Merge the latest release into dev and push upstream.

          make release-merge-stable-to-next RELEASE_PREVIOUS=release_${previous_version}
          make release-push-dev

    - [ ] Create and push release branch:

          make release-create-rc

    - [ ] Open pull requests from your fork of branch ``version-${version}`` to upstream ``release_${version}`` and of ``version-${next_version}.dev`` to ``dev``.
    - [ ] Update ``MILESTONE_NUMBER`` in the [maintenance bot](https://github.com/galaxyproject/galaxy/blob/dev/.github/workflows/maintenance_bot.yaml) to `${next_version}` so it properly tags new pull requests.

- [ ] **Issue Review Timeline Notes**

    - [ ] Ensure any security fixes will be ready prior to ${freeze_date} + 1 week, to allow time for notification prior to release.
    - [ ] Ensure ownership of outstanding bugfixes and track progress during freeze.

- [ ] **Deploy and Test Release**

    - [ ] Update test.galaxyproject.org to ensure it is running a dev at or past branch point (${freeze_date} + 1 day).
    - [ ] Update testtoolshed.g2.bx.psu.edu to ensure it is running a dev at or past branch point (${freeze_date} + 1 day).
    - [ ] Deploy to usegalaxy.org (${freeze_date} + 1 week).
    - [ ] Deploy to toolshed.g2.bx.psu.edu (${freeze_date} + 1 week).
    - [ ] [Update BioBlend CI testing](https://github.com/galaxyproject/bioblend/blob/main/.github/workflows/test.yaml) to include a ``release_${version}`` target: add ``- release_${version}`` to the ``galaxy_version`` list in ``.github/workflows/test.yaml`` .
    - [ ] Update GALAXY_RELEASE in IUC and devteam github workflows
        - [ ] https://github.com/galaxyproject/tools-iuc/blob/master/.github/workflows/
        - [ ] https://github.com/galaxyproject/tools-devteam/blob/master/.github/workflows/

- [ ] **Create Release Notes**

    - [ ] Review merged pull requests and ensure they all have a milestone attached. [Link](https://github.com/galaxyproject/galaxy/pulls?utf8=%E2%9C%93&q=is%3Apr+is%3Amerged+no%3Amilestone+-label%3Amerge+)
    - [ ] Switch to release branch and create a new branch for release notes

          git checkout release_${version} -b ${version}_release_notes
    - [ ] Bootstrap the release notes

          make release-bootstrap-history RELEASE_CURR=${version}
    - [ ] Open newly created files and manually curate major topics and release notes.
    - [ ] Run ``python scripts/release-diff.py release_${previous_version}`` and add configuration changes to release notes.
    - [ ] Add new release to doc/source/releases/index.rst
    - [ ] Open a pull request for the release notes branch.
    - [ ] Merge release note pull request.

- [ ] **Do Release**

    - [ ] Ensure all [blocking milestone issues](https://github.com/galaxyproject/galaxy/issues?q=is%3Aopen+is%3Aissue+milestone%3A${version}) have been resolved.

          make release-check-blocking-issues RELEASE_CURR=${version}
    - [ ] Ensure all [blocking milestone pull requests](https://github.com/galaxyproject/galaxy/pulls?q=is%3Aopen+is%3Apr+milestone%3A${version}) have been merged or closed.

          make release-check-blocking-prs RELEASE_CURR=${version}
    - [ ] Ensure all pull requests merged into the pre-release branch during the freeze have [milestones attached](https://github.com/galaxyproject/galaxy/pulls?q=is%3Apr+is%3Aclosed+base%3Arelease_${version}+is%3Amerged+no%3Amilestone) and that they are the not [${next_version} milestones](https://github.com/galaxyproject/galaxy/pulls?q=is%3Apr+is%3Aclosed+base%3Arelease_${version}+is%3Amerged+milestone%3A${next_version})
    - [ ] Ensure release notes include all pull requests added during the freeze by re-running the release note bootstrapping:

          make release-bootstrap-history
    - [ ] Ensure previous release is merged into current. [GitHub branch comparison](https://github.com/galaxyproject/galaxy/compare/release_${version}...release_${previous_version})
    - [ ] Create and push release tag:

          make release-create

    - [ ] Create dev packages:

          cd packages && ./build_packages.sh

    - [ ] Create the first point release (v${version}.0) using the instructions at https://docs.galaxyproject.org/en/master/dev/create_point_release.html
    - [ ] Open PR against planemo with a pin to the new packages

- [ ] **Do Docker Release**

    - [ ] Change the [dev branch](https://github.com/bgruening/docker-galaxy-stable/tree/dev) of the Galaxy Docker container to ${next_version}
    - [ ] Merge dev into master

- [ ] **Announce Release**

    - [ ] Verify release included in https://docs.galaxyproject.org/en/master/releases/index.html
    - [ ] Review announcement in https://github.com/galaxyproject/galaxy/blob/dev/doc/source/releases/${version}_announce.rst
    - [ ] Stage announcement content (Hub, Galaxy Help, etc.) on announce date to capture date tags. Note: all final content does not need to be completed to do this.
    - [ ] Create hub *highlights* and post as a new "news" content item. [An example](https://galaxyproject.org/news/2018-9-galaxy-release/).
    - [ ] Tweet docs news *highlights* link as @galaxyproject on twitter. [An example](https://twitter.com/galaxyproject/status/973646125633695744).
    - [ ] Post *highlights* with tags `news` and `release` to [Galaxy Help](https://help.galaxyproject.org/). [An example](https://help.galaxyproject.org/t/galaxy-release-19-01/712).
    - [ ] Email *highlights* to [galaxy-dev](http://dev.list.galaxyproject.org/) and [galaxy-announce](http://announce.list.galaxyproject.org/) @lists.galaxyproject.org. [An example](http://dev.list.galaxyproject.org/The-Galaxy-release-16-04-is-out-tp4669419.html)
    - [ ] Adjust http://getgalaxy.org text and links to match current master branch by opening a PR at https://github.com/galaxyproject/galaxy-hub/

- [ ] **Prepare for next release**

    - [ ] Close milestone ``${version}`` and ensure milestone ``${next_version}`` exists.
    - [ ] Create release issue for next version ``make release-issue``.
    - [ ] Schedule committer meeting to discuss re-alignment of priorities.
    - [ ] Close this issue.
"""  # noqa: E501
)

release_version_argument = click.argument("release-version", type=ClickVersion())


@click.group(help="Subcommands of this script can perform various tasks around creating Galaxy releases")
def cli():
    pass


@cli.command(help="Create release checklist issue on GitHub")
@group_options(release_version_argument, galaxy_root_option)
def create_release_issue(release_version: Version, galaxy_root: Path):
    previous_release = _previous_release(galaxy_root, release_version)
    new_version_params = _next_version_params(release_version)
    next_version = new_version_params["version"]
    freeze_date, _ = _release_dates(release_version)
    release_issue_template_params = dict(
        version=release_version,
        next_version=next_version,
        previous_version=previous_release,
        freeze_date=freeze_date,
    )
    release_issue_contents = RELEASE_ISSUE_TEMPLATE.safe_substitute(**release_issue_template_params)
    github = github_client()
    repo = github.get_repo(f"{PROJECT_OWNER}/{PROJECT_NAME}")
    release_issue = repo.create_issue(
        title=f"Publication of Galaxy Release v {release_version}",
        body=release_issue_contents,
    )
    return release_issue


@cli.command(help="Create or update release changelog")
@group_options(release_version_argument, galaxy_root_option)
def create_changelog(release_version: Version, galaxy_root: Path):
    release_file = _release_file(galaxy_root, str(release_version) + ".rst")
    enhancement_targets = "\n\n".join(f".. enhancement_tag_{a}" for a in GROUPED_TAGS.values())
    bug_targets = "\n\n".join(f".. bug_tag_{a}" for a in GROUPED_TAGS.values())
    template = TEMPLATE
    template = template.replace(".. enhancement", f"{enhancement_targets}\n\n.. enhancement")
    template = template.replace(".. bug", f"{bug_targets}\n\n.. bug")
    release_info = string.Template(template).safe_substitute(release=release_version)
    _write_file(release_file, release_info, skip_if_exists=True)
    month = MINOR_TO_MONTH[release_version.minor]
    month_name = calendar.month_name[month]
    year = release_version.major

    announce_info = ANNOUNCE_TEMPLATE.substitute(month_name=month_name, year=year, release=release_version)
    announce_file = _release_file(galaxy_root, str(release_version) + "_announce.rst")
    _write_file(announce_file, announce_info, skip_if_exists=True)

    announce_user_info = ANNOUNCE_USER_TEMPLATE.substitute(month_name=month_name, year=year, release=release_version)
    announce_user_file = _release_file(galaxy_root, str(release_version) + "_announce_user.rst")
    _write_file(announce_user_file, announce_user_info, skip_if_exists=True)

    prs_file = _release_file(galaxy_root, str(release_version) + "_prs.rst")
    seen_prs = set()
    try:
        with open(prs_file) as fh:
            seen_prs = set(map(int, re.findall(r"\.\. _Pull Request (\d+): https", fh.read())))
    except FileNotFoundError:
        pass
    _write_file(prs_file, PRS_TEMPLATE, skip_if_exists=True)

    next_version_params = _next_version_params(release_version)
    next_version = next_version_params["version"]
    next_release_file = _release_file(galaxy_root, str(next_version) + "_announce.rst")

    next_announce = NEXT_TEMPLATE.substitute(**next_version_params)
    with open(next_release_file, "w") as fh:
        fh.write(next_announce)
    releases_index = _release_file(galaxy_root, "index.rst")
    releases_index_contents = _read_file(releases_index)
    releases_index_contents = releases_index_contents.replace(
        ".. announcements\n",
        ".. announcements\n   " + str(next_version) + "_announce\n",
    )
    _write_file(releases_index, releases_index_contents, skip_if_exists=True)

    for pr in _get_prs(str(release_version)):
        if pr.number not in seen_prs:
            pr_to_doc(
                galaxy_root=galaxy_root,
                release_version=release_version,
                pr=pr,
            )


@cli.command(help="List release blocking PRs")
@group_options(release_version_argument)
def check_blocking_prs(release_version: Version):
    block = False
    for pr in _get_prs(str(release_version), state="open"):
        click.echo(f"Blocking PR| {_pr_to_str(pr)}", err=True)
        block = True
    if block:
        sys.exit(1)


@cli.command(help="List release blocking issues")
@group_options(release_version_argument)
def check_blocking_issues(release_version: Version):
    block = 0
    github = github_client()
    repo = github.get_repo(f"{PROJECT_OWNER}/{PROJECT_NAME}")
    issues = repo.get_issues(state="open")
    for issue in issues:
        # issue can also be a pull request, which could be filtered out with `not issue.pull_request`
        if (
            issue.milestone
            and issue.milestone.title == str(release_version)
            and "Publication of Galaxy Release" not in issue.title
        ):
            click.echo(f"Blocking issue| {_issue_to_str(issue)}", err=True)
            block = 1

    sys.exit(block)


def _issue_to_str(pr):
    if isinstance(pr, str):
        return pr
    return f"Issue #{pr.number} ({pr.title}) {pr.html_url}"


def _next_version_params(release_version: Version):
    # we'll just hardcode this to 3 "minor" versions per year
    if release_version.minor < 2:
        next_major = release_version.major
        next_minor = release_version.minor + 1
    else:
        next_major = release_version.major + 1
        next_minor = 0
    next_month_name = calendar.month_name[MINOR_TO_MONTH[next_minor]]
    next_version = Version(f"{next_major}.{next_minor}")
    freeze_date, release_date = _release_dates(next_version)
    return dict(
        version=next_version,
        year=next_major,
        month_name=next_month_name,
        freeze_date=freeze_date,
        release_date=release_date,
    )


def _release_dates(version: Version):
    # hardcoded to 3 releases a year, freeze dates kind of random
    year = version.major
    month = MINOR_TO_MONTH[version.minor]
    first_of_month = datetime.date(year + 2000, month, 1)
    freeze_date = next_weekday(first_of_month, 0)
    release_date = next_weekday(first_of_month, 0) + datetime.timedelta(21)
    return freeze_date, release_date


def _get_prs(release_version: str, state="closed"):
    github = github_client()
    repo = github.get_repo(f"{PROJECT_OWNER}/{PROJECT_NAME}")
    pull_requests = repo.get_pulls(state=state)
    reached_old_prs = False

    for pr in pull_requests:
        if reached_old_prs:
            break

        if pr.created_at.replace(tzinfo=None) < datetime.datetime(2020, 5, 1, 0, 0):
            reached_old_prs = True
            pass
        merged_at = pr.merged_at
        milestone = pr.milestone
        proper_state = state != "closed" or merged_at
        if not proper_state or not milestone or milestone.title != release_version:
            continue
        yield pr


def pr_to_doc(galaxy_root: Path, release_version: Version, pr: PullRequest):
    history_path = _release_file(galaxy_root, str(release_version)) + ".rst"
    user_announce_path = history_path[0 : -len(".rst")] + "_announce_user.rst"
    prs_path = history_path[0 : -len(".rst")] + "_prs.rst"

    history = _read_file(history_path)
    user_announce = _read_file(user_announce_path)
    prs_content = _read_file(prs_path)

    def extend_target(target, line, source=history):
        from_str = f".. {target}\n"
        if target not in source:
            raise Exception(f"Failed to find target [{target}] in source [{source}]")
        return source.replace(from_str, from_str + line + "\n")

    text_target = "to_doc"
    to_doc = pr.title.rstrip(".") + " "

    owner = None
    user = pr.user
    owner = user.login
    text = f".. _Pull Request {pr.number}: {PROJECT_URL}/pull/{pr.number}"
    prs_content = extend_target("github_links", text, prs_content)
    to_doc += f"\n(thanks to `@{owner} <https://github.com/{owner}>`__)."
    to_doc += f"\n`Pull Request {pr.number}`_"
    labels = _pr_to_labels(pr)
    text_target = _text_target(pr)

    to_doc = wrap(to_doc)
    if text_target is not None:
        history = extend_target(text_target, to_doc, history)
    if "area/datatypes" in labels:
        user_announce = extend_target("datatypes", to_doc, user_announce)
    if "area/visualizations" in labels:
        user_announce = extend_target("visualizations", to_doc, user_announce)
    if "area/tools" in labels:
        user_announce = extend_target("tools", to_doc, user_announce)
    _write_file(history_path, history)
    _write_file(prs_path, prs_content)
    _write_file(user_announce_path, user_announce)


def _read_file(path):
    with open(path) as f:
        return f.read()


def _write_file(path, contents, skip_if_exists=False):
    if skip_if_exists and os.path.exists(path):
        return
    with open(path, "w") as f:
        f.write(contents)


def _previous_release(galaxy_root, to: Version):
    previous_release = None
    for release in _releases(galaxy_root):
        if release == str(to):
            break

        previous_release = release

    return previous_release


def _releases(galaxy_root):
    releases_path = galaxy_root / "doc" / "source" / "releases"
    all_files = sorted(os.listdir(releases_path))
    release_note_file_pattern = re.compile(r"\d+\.\d+.rst")
    release_note_files = [f for f in all_files if release_note_file_pattern.match(f)]
    return sorted(f.rstrip(".rst") for f in release_note_files)


def _release_file(galaxy_root: Path, release: Optional[str]) -> str:
    releases_path = galaxy_root / "doc" / "source" / "releases"
    if release is None:
        release = sorted(os.listdir(releases_path))[-1]
    history_path = os.path.join(releases_path, release)
    return history_path


def get_first_sentence(message: str) -> str:
    first_line = message.split("\n")[0]
    return first_line


def process_sentence(message):
    # Strip tags like [15.07].
    message = strip_release(message=message)
    # Link issues and pull requests...
    issue_url = f"https://github.com/{PROJECT_OWNER}/{PROJECT_NAME}/issues"
    message = re.sub(r"#(\d+)", rf"`#\1 <{issue_url}/\1>`__", message)
    return message


def wrap(message):
    message = process_sentence(message)
    wrapper = textwrap.TextWrapper(initial_indent="* ")
    wrapper.subsequent_indent = "  "
    wrapper.width = 160
    message_lines = message.splitlines()
    first_lines = "\n".join(wrapper.wrap(message_lines[0]))
    wrapper.initial_indent = "  "
    rest_lines = "\n".join("\n".join(wrapper.wrap(m)) for m in message_lines[1:])
    return first_lines + ("\n" + rest_lines if rest_lines else "")


def next_weekday(d, weekday):
    """Return the next week day (0 for Monday, 6 for Sunday) starting from ``d``."""
    days_ahead = weekday - d.weekday()
    if days_ahead <= 0:  # Target day already happened this week
        days_ahead += 7
    return d + datetime.timedelta(days_ahead)



import pathlib
from typing import (
    Any,
    Optional,
)

import click
from click.core import (
    Context,
    Parameter,
)
from packaging.version import Version

galaxy_root_option = click.option(
    "--galaxy-root",
    type=click.Path(exists=True, file_okay=False, resolve_path=True, path_type=pathlib.Path),
    default=".",
)


def group_options(*options):
    def wrapper(function):
        for option in reversed(options):
            function = option(function)
        return function

    return wrapper


class ClickVersion(click.ParamType):
    name = "pep440 version"

    def convert(self, value: Any, param: Optional[Parameter], ctx: Optional[Context]) -> Version:
        try:
            return Version(value)
        except Exception as e:
            self.fail(f"{value!r} is not a valid PEP440 version number: {str(e)}", param, ctx)



import json
import os

from github import Github


def github_client() -> Github:
    """Search environment for github access token and produce client object.

    Easiest thing to do is just to set an environment variable GITHUB_AUTH to
    your personal access token (
    https://docs.github.com/en/authentication/keeping-your-account-and-data-secure/creating-a-personal-access-token
    ). Alternatively, this can be placed into a json file in ~/.github.json in a
    map keyed on login_or_token.
    """
    auth = os.environ.get("GITHUB_AUTH")
    if auth is not None:
        return Github(auth)
    else:
        github_json_path = os.path.expanduser("~/.github.json")
        if not os.path.exists(github_json_path):
            return Github(None)
        with open(github_json_path) as fh:
            github_json_dict = json.load(fh)
        return Github(**github_json_dict)



import re
from typing import List

from github.PullRequest import PullRequest

PROJECT_OWNER = "galaxyproject"
PROJECT_NAME = "galaxy"
PROJECT_URL = f"https://github.com/{PROJECT_OWNER}/{PROJECT_NAME}"
PROJECT_API = f"https://api.github.com/repos/{PROJECT_OWNER}/{PROJECT_NAME}/"

GROUPED_TAGS = dict(
    [
        ("area/visualizations", "viz"),
        ("area/datatypes", "datatypes"),
        ("area/tools", "tools"),
        ("area/workflows", "workflows"),
        ("area/client", "ui"),
        ("area/jobs", "jobs"),
        ("area/admin", "admin"),
    ]
)


def _pr_to_str(pr):
    if isinstance(pr, str):
        return pr
    return f"PR #{pr.number} ({pr.title}) {pr.html_url}"


def _text_target(pull_request: PullRequest, skip_merge=True):
    pr_number = pull_request.number
    labels = [label.name.lower() for label in pull_request.labels]
    is_bug = is_enhancement = is_feature = is_minor = is_major = is_merge = is_small_enhancement = False
    if len(labels) == 0:
        print(f"No labels found for {pr_number}")
        return None
    for label_name in labels:
        if label_name == "minor":
            is_minor = True
        elif label_name == "major":
            is_major = True
        elif label_name == "merge":
            is_merge = True
        elif label_name == "kind/bug":
            is_bug = True
        elif label_name == "kind/feature":
            is_feature = True
        elif label_name == "kind/enhancement":
            is_enhancement = True
        elif label_name in ["kind/testing", "kind/refactoring"]:
            is_small_enhancement = True
        elif label_name == "procedures":
            # Treat procedures as an implicit enhancement.
            is_enhancement = True

    is_some_kind_of_enhancement = is_enhancement or is_feature or is_small_enhancement

    if not (is_bug or is_some_kind_of_enhancement or is_minor or is_merge):
        print(f"No 'kind/*' or 'minor' or 'merge' or 'procedures' label found for {_pr_to_str(pull_request)}")
        text_target = None

    if is_minor or is_merge and skip_merge:
        return

    if is_some_kind_of_enhancement and is_major:
        text_target = "major_feature"
    elif is_feature:
        text_target = "feature"
    elif is_enhancement:
        for label, tag in GROUPED_TAGS.items():
            if label in labels:
                text_target = f"enhancement_tag_{tag}"
                break
        else:
            text_target = "enhancement"
    elif is_some_kind_of_enhancement:
        text_target = "small_enhancement"
    elif is_major:
        text_target = "major_bug"
    elif is_bug:
        for label, tag in GROUPED_TAGS.items():
            if label in labels:
                text_target = f"bug_tag_{tag}"
                break
        else:
            text_target = "bug"
    else:
        print(f"Logic problem, cannot determine section for {_pr_to_str(pull_request)}")
        text_target = None
    if text_target:
        text_target += "\n"
    return text_target


def _pr_to_labels(pr: PullRequest) -> List[str]:
    labels = [label.name.lower() for label in pr.labels]
    return labels


def strip_release(message):
    return re.sub(r"^\s*\[.*\]\s*", r"", message)



import pathlib

import pytest

from galaxy_release_util.point_release import (
    get_next_devN_version,
    get_root_version,
    get_sorted_package_paths,
)

VERSION_PY_CONTENTS = """VERSION_MAJOR = "23.0"
VERSION_MINOR = "2"
VERSION = VERSION_MAJOR + (f".{VERSION_MINOR}" if VERSION_MINOR else "")
"""

PACKAGES_BY_DEP_DAG_CONTENTS = """
foo

bar
#this is a comment
baz
"""


def write_contents(path: pathlib.Path, contents: str):
    path.parent.mkdir(parents=True)
    path.write_text(contents)


@pytest.fixture()
def galaxy_root(tmp_path: pathlib.Path):
    write_contents(tmp_path / "lib" / "galaxy" / "version.py", VERSION_PY_CONTENTS)
    write_contents(tmp_path / "packages" / "packages_by_dep_dag.txt", PACKAGES_BY_DEP_DAG_CONTENTS)
    return tmp_path


def test_get_root_version(galaxy_root):
    version = get_root_version(galaxy_root)
    assert version.major == 23
    assert version.minor == 0
    assert version.micro == 2


def test_get_next_devN_version(galaxy_root):
    version = get_next_devN_version(galaxy_root)
    assert version.is_devrelease
    assert version.major == 23
    assert version.minor == 0
    assert version.micro == 3
    assert version.dev == 0


def test_get_sorted_package_paths(galaxy_root):
    packages = get_sorted_package_paths(galaxy_root)
    assert len(packages) == 3
    assert packages[0].name == "foo"
    assert packages[1].name == "bar"
    assert packages[2].name == "baz"



import click

from ..bootstrap_history import cli as boostrap_cli
from ..point_release import cli as point_release_cli

cli = click.CommandCollection(
    sources=[boostrap_cli, point_release_cli],
    help="Perform various tasks around creating Galaxy releases and point releases",
)
